@@ -33,13 +33,6 @@ class TimestampNotAvailableException(Exception):
     pass
 
 
-class InvalidAssignByKeyException(Exception):
-    """
-    Exception to raise when consumer assigned by an invalid option
-    """
-    pass
-
-
 class ProtobufCompilerException(Exception):
     """
     Exception to raise when protobuf compiler fails
